(*----------------------------------------------------------------------------*
 *  Direct3D GifBlender sample                                                *
 *  demostrate how to load and render GIF file.                               *
 *  This sample uses Ander's Melander TGifImage component internally to handle*
 *  GIF animation. Then GIF image is updated by TGifImage component - image   *
 *  is loaded to standart Direct3D texture someone can use for shading.       *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 03-Oct-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit GifBlenderUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, mmsystem, Graphics, GifImage,
  D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil,
  {$IFDEF DXG_COMPAT} DirectXGraphics {$ELSE} Direct3D8 {$ENDIF};

type
  TCustomVertex = record
    position: TD3DXVector3;   // The position
    color: TD3DColor;         // The color
    tu,tv: Single;            // The texture coordinates
  end;
  PCustomVertex = array[0..0] of TCustomVertex;

const
  // Our custom FVF, which describes our custom vertex structure
  D3DFVF_CUSTOMVERTEX = (D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1);

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pVB: IDirect3DVertexBuffer8; // Buffer to hold vertices
    m_pTexture: IDirect3DTexture8; // Our texture

    m_pFont: CD3DFont;                        // Font for drawing text
    m_pFontSmall: CD3DFont;                   // Font for drawing text

    GIF: TGIFImage;
    m_Bitmap: TBitmap;
    m_CurrentFrame: Integer;
    m_NeedToUpdateTexture: Boolean;

    function InitGeometry: HRESULT;
    procedure SetupNextFrame;
    procedure OnAnimate(Sender: TObject);
  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; override;

  public
    constructor Create;
  end;

implementation

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle    := 'GifBlender: painting Gif''s in Direct3D';
  m_bUseDepthBuffer   := True;

  m_pFont           := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall      := CD3DFont.Create('Arial', 9, D3DFONT_BOLD);
  m_pVB             := nil;

  // Load the GIF that will be converted to texture images
  GIF := TGIFImage.Create;
  GIF.LoadFromFile(ExtractFilePath(ParamStr(0))+'star.gif');
  GIF.OnPaint:= OnAnimate;
  m_CurrentFrame:= -1;
  m_NeedToUpdateTexture:= True;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit;
begin
  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  matWorld: TD3DXMatrix;
  matView: TD3DXMatrix;
  matProj: TD3DXMatrix;
begin
  Result:= S_OK;

  // For our world matrix, we will just leave it as the identity
  D3DXMatrixIdentity(matWorld);
  D3DXMatrixRotationX(matWorld, timeGetTime/1000.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  // Set up our view matrix. A view matrix can be defined given an eye point,
  // a point to lookat, and a direction for which way is up. Here, we set the
  // eye five units back along the z-axis and up three units, look at the
  // origin, and define "up" to be in the y-direction.
  D3DXMatrixLookAtLH(matView, D3DXVector3(0.0, 3.0,-5.0),
                              D3DXVector3(0.0, 0.0, 0.0),
                              D3DXVector3(0.0, 1.0, 0.0));
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  // For the projection matrix, we set up a perspective transform (which
  // transforms geometry from 3D view space to 2D viewport space, with
  // a perspective divide making objects smaller in the distance). To build
  // a perpsective transform, we need the field of view (1/4 pi is common),
  // the aspect ratio, and the near and far clipping planes (which define at
  // what distances geometry should be no longer be rendered).
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, 1.0, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------

function CMyD3DApplication.Render;
{$IFDEF SHOW_HOW_TO_USE_TCI}
var
  mat: TD3DXMatrix;
{$ENDIF}
begin
  // Clear the backbuffer and the zbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,255), 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Here we update and assign texture to GIF current frame
    if m_NeedToUpdateTexture then SetupNextFrame;

    // Setup our texture. Using textures introduces the texture stage states,
    // which govern how textures get blended together (in the case of multiple
    // textures) and lighting information. In this case, we are modulating
    // (blending) our texture with the diffuse color of the vertices.
    m_pd3dDevice.SetTexture(0, m_pTexture);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);

  {$IFDEF SHOW_HOW_TO_USE_TCI}
    // Note: to use D3D texture coordinate generation, use the stage state
    // D3DTSS_TEXCOORDINDEX, as shown below. In this example, we are using
    // the position of the vertex in camera space to generate texture
    // coordinates. The tex coord index (TCI) parameters are passed into a
    // texture transform, which is a 4x4 matrix which transforms the x,y,z
    // TCI coordinates into tu, tv texture coordinates.

    // In this example, the texture matrix is setup to
    // transform the texture from (-1,+1) position coordinates to (0,1)
    // texture coordinate space:
    //    tu =  0.5*x + 0.5
    //    tv = -0.5*y + 0.5

    mat._11 := 0.25; mat._12 := 0.00; mat._13 := 0.00; mat._14 := 0.00;
    mat._21 := 0.00; mat._22 :=-0.25; mat._23 := 0.00; mat._24 := 0.00;
    mat._31 := 0.00; mat._32 := 0.00; mat._33 := 1.00; mat._34 := 0.00;
    mat._41 := 0.50; mat._42 := 0.50; mat._43 := 0.00; mat._44 := 1.00;

    m_pd3dDevice.SetTransform(D3DTS_TEXTURE0, mat);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION);
  {$ENDIF}

    // Render the vertex buffer contents
    m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TCustomVertex));
    m_pd3dDevice.SetVertexShader(D3DFVF_CUSTOMVERTEX);
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2*50-2);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // Report current frame of GIF
    m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255),
      PChar('Current GIF frame: ' + IntToStr(m_CurrentFrame)));

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Creates the scene geometry
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitGeometry: HRESULT;
var
  pVertices: ^PCustomVertex;
  i: DWORD;
  theta: Single;
begin
  Result:= E_FAIL;

  // Create GIF and memory-bitmap container 
  m_Bitmap:= TBitmap.Create;
  m_Bitmap.Assign(GIF);

  // Start painting GIF to memory-bitmap
  GIF.Paint(m_Bitmap.Canvas, Rect(0,0,GIF.Width,GIF.Height),
            GIFImageDefaultDrawOptions);

  // Create the vertex buffer.
  if FAILED(m_pd3dDevice.CreateVertexBuffer(50*2*SizeOf(TCustomVertex),
                                            0, D3DFVF_CUSTOMVERTEX,
                                            D3DPOOL_DEFAULT, m_pVB))
  then Exit;

  // Fill the vertex buffer. We are setting the tu and tv texture
  // coordinates, which range from 0.0 to 1.0
  if FAILED(m_pVB.Lock(0, 0, PByte(pVertices), 0))
  then Exit;

  for i:= 0 to 49 do
  begin
    theta := (2*D3DX_PI*i)/(50-1);
    pVertices[2*i+0].position := D3DXVector3(sin(theta),-1.0, cos(theta));
    pVertices[2*i+0].color    := $ffffffff;
    pVertices[2*i+0].tu       := i/(50-1);
    pVertices[2*i+0].tv       := 1.0;

    pVertices[2*i+1].position := D3DXVector3(sin(theta), 1.0, cos(theta));
    pVertices[2*i+1].color    := $ff808080;
    pVertices[2*i+1].tu       := i/(50-1);
    pVertices[2*i+1].tv       := 0.0;
  end;

  m_pVB.Unlock;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
begin
  // Restore the font
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
var
  matIdentity: TD3DXMatrix;
  matProj: TD3DXMatrix;
  fAspect: Single;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  Result:= InitGeometry;
  if FAILED(Result) then Exit;

  // Turn off culling
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  // Turn off D3D lighting
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);

  // Turn on the zbuffer
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the projection matrix
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  m_pVB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  m_pTexture:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pFontSmall);
  FreeAndNil(GIF);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord;
  Format: TD3DFormat): HResult;
begin
  if dwBehavior and D3DCREATE_PUREDEVICE <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

procedure CMyD3DApplication.SetupNextFrame;
var
  S: TMemoryStream;
begin
  S:= TMemoryStream.Create;
  try
    m_Bitmap.SaveToStream(S);
    D3DXCreateTextureFromFileInMemory(m_pd3dDevice, S.Memory^, S.Size, m_pTexture);
    m_NeedToUpdateTexture:= False;
  finally
    S.Free;
  end;
end;

// Here we notify our main program when to update texture with new GIF frame
procedure CMyD3DApplication.OnAnimate(Sender: TObject);
begin
  if not(Sender is TGIFPainter) then Exit;

  m_CurrentFrame:= TGIFPainter(Sender).ActiveImage;
  m_NeedToUpdateTexture:= True;
end;

end.

